@@ -6,7 +6,7 @@ use Carp qw/croak/;
 use XSLoader;
 use Sereal::Path::Iterator;
 
-our $VERSION    = '0.013';
+our $VERSION    = '0.014';
 our $XS_VERSION = $VERSION; $VERSION= eval $VERSION;
 
 XSLoader::load(__PACKAGE__, $Sereal::Path::VERSION);
@@ -120,13 +120,21 @@ authors would like to express their gratitude.
 
 Ivan Kruglov <ivan.kruglov@yahoo.com>
 
+=head1 CONTRIBUTORS
+
+Roman Studenikin <roman.studenikin@booking.com>
+
+Steven Lee <stevenwh.lee@gmail.com>
+
+Gonzalo Diethem <gonzalo.diethelm@gmail.com>
+
 =head1 COPYRIGHT AND LICENCE
 
 Copyright 2007 Stefan Goessner.
 
 Copyright 2010-2013 Toby Inkster.
 
-Copyright 2014-2015 Ivan Kruglov.
+Copyright 2014-2016 Ivan Kruglov.
 
 This module is tri-licensed. It is available under the X11 (a.k.a. MIT)
 licence; you can also redistribute it and/or modify it under the same
